﻿-- Author      : RisM
-- Create Date : 6/28/2009 3:56:24 PM

SpeakinSpell = LibStub("AceAddon-3.0"):GetAddon("SpeakinSpell")
local L	= LibStub("AceLocale-3.0"):GetLocale("SpeakinSpell", false)
local DEFAULT_SPEECHES = LibStub("AceLocale-3.0"):GetLocale("SpeakinSpell_DEFAULT_SPEECHES",	false)


-------------------------------------------------------------------------------
-- DATA STRUCTURES (? not sure what to call this category)
-------------------------------------------------------------------------------


-------------------------------------------------------------------------------
-- INITIALIZATION / IMPORT OF DEFAULT DATA
-------------------------------------------------------------------------------


function SpeakinSpell:InitDefaultSavedData()
	-- copy defaults into current settings using class-specific defaults
	local playerClass, englishClass = UnitClass("player"); --englishClass is the UPPERCASE class name in English
	local playerRace, englishRace = UnitRace("player"); 
	englishRace = string.upper(englishRace) -- for table keys
	
	SpeakinSpellSavedData.Version = SpeakinSpell.CURRENT_VERSION
	SpeakinSpellSavedData.ShowSpellEvents = false
	SpeakinSpellSavedData.ShowDebugMessages = false
	SpeakinSpellSavedData.EnableAllMessages = true
	SpeakinSpellSavedData.ShowAllRanks = false
	
	SpeakinSpellSavedData.EventTable = nil
	SpeakinSpellSavedData.EventTable = {}

	-- Insert defaults speeches and event settings for: All, Race, and Class defaults
	self:AddTable( SpeakinSpellSavedData.EventTable, DEFAULT_SPEECHES.ALL )
	self:AddTable( SpeakinSpellSavedData.EventTable, DEFAULT_SPEECHES.RACES[englishRace] )
	self:AddTable( SpeakinSpellSavedData.EventTable, DEFAULT_SPEECHES.CLASSES[englishClass] )
	
	-- Reset the new events table
	SpeakinSpellSavedData.NewEventsDetected = nil
	SpeakinSpellSavedData.NewEventsDetected = {}

	-- remove faction-specific spells (i.e. mage portals) 
	-- and reset runtime data such as localizable display names
	for key, EventTableEntry in pairs(SpeakinSpellSavedData.EventTable) do
		-- rebuild some info that's normally built dynamically
		local de = EventTableEntry.DetectedEvent
		de.DisplayName = self:FormatDisplayName(de)
		de.key = key
		de.SpellLink = self:SafeGetSpellLink(de)
		
		if EventTableEntry.Faction ~= nil then
			if EventTableEntry.Faction ~= UnitFactionGroup("player") then
				SpeakinSpellSavedData.EventTable[key] = nil
			end
		end
	end
	
	-- create the random substitutions tables
	SpeakinSpellSavedData.RandomSubs = nil
	SpeakinSpellSavedData.RandomSubs = {}
	self:AddTable( SpeakinSpellSavedData.RandomSubs, DEFAULT_SPEECHES.RandomSubs )
	-- Add guild and arena teams to the <randomfaction> substitution
	-- NOTE: don't do this during InitRuntimeData because that is also done when resetting defaults
	--		and we only need this to happen once after loading the addon into memory
	self:AddRandomFactions()
	

end


function SpeakinSpell:ImportNewSpeechesFor(key, Messages)
	local funcname = "ImportNewSpeechesFor"
	-- add any new messages from the defaults into our current settings
	-- don't import duplicates
	-- and don't overwrite by numeric index keys like CopyTable or AddTable would do

	-- we will copy from Messages into Target
	local Target = SpeakinSpellSavedData.EventTable[key].Messages

	local ShowedName = false

	for _,newmsg in pairs(Messages) do
		-- see if we already use this message
		local found = false
		for _,oldmsg in pairs(Target) do
			if oldmsg == newmsg then
				found = true
			end
		end

		if not found then
			-- this is a new message

			-- use the chat window to report what we're adding
			-- TODO: add prompts
			if not ShowedName then
				local de = SpeakinSpellSavedData.EventTable[key].DetectedEvent
				de.DisplayName = self:FormatDisplayName(de)
				self:Print(de.DisplayName)
				ShowedName = true
			end

			--self:Print(newmsg)
			tinsert(Target, newmsg)
		end
	end
end


function SpeakinSpell:ImportNewDefaults()
	local funcname = "ImportNewDefaults"

	local playerClass, englishClass = UnitClass("player"); --englishClass is the UPPERCASE class name in English
	local playerRace, englishRace = UnitRace("player");
	englishRace = string.upper(englishRace) -- for table keys

	self:DebugMsgDumpString("playerRace",playerRace)
	self:DebugMsgDumpString("englishRace",englishRace)

	NEW_DEF_MSG_SOURCES = {
		DEFAULT_SPEECHES.ALL,
		DEFAULT_SPEECHES.RACES[englishRace],
		DEFAULT_SPEECHES.CLASSES[englishClass],
	}

	self:Print(L["Adding new default samples to your saved settings"])

	for _,NewEventTable in pairs(NEW_DEF_MSG_SOURCES) do
		for key, NewEvent in pairs(NewEventTable) do
			self:DebugMsg(funcname, "key:"..key)
			if (NewEvent.Faction == nil) or (NewEvent.Faction == UnitFactionGroup("player")) then 
				-- the new event applies to our faction so use it
				if SpeakinSpellSavedData.EventTable[key] then
					-- we already have settings for this event, so merge them
					-- add any new messages from the defaults into our current settings
					self:ImportNewSpeechesFor(key, NewEvent.Messages)
				else
					-- we don't have settings for this event
					-- add a copy of this new default event to the current settings
					-- TODO: add prompts
					SpeakinSpellSavedData.EventTable[key] = self:CopyTable(NewEventTable[key])
					
					-- validate and display the DisplayName
					local de = SpeakinSpellSavedData.EventTable[key].DetectedEvent
					de.DisplayName = self:FormatDisplayName(de)
					self:Print(de.DisplayName)
					--self:Print(L["(new event - all speeches are new)"])
				end
			end
		end
	end

	-- import random substitutions
	-- TODO: add prompts
	self:AddTable( SpeakinSpellSavedData.RandomSubs, DEFAULT_SPEECHES.RandomSubs )
	-- Add guild and arena teams to the <randomfaction> substitution
	-- NOTE: don't do this during InitRuntimeData because that is also done when resetting defaults
	--		and we only need this to happen once after loading the addon into memory
	-- TODO: add prompts
	self:AddRandomFactions()

	self:Print(L["Import complete"])
end


-------------------------------------------------------------------------------
-- DATA STRUCTURES (? not sure what to call this category)
-------------------------------------------------------------------------------

function SpeakinSpell:InitRuntimeData()
	-- this is just a declaration
	self.RuntimeData = {
		AnnouncementHistory = {}, -- [eventkey] = {LastMessage, LastTime, LastTarget}
		AnnouncedThisCombat = {}, -- [eventkey] = true, for Limit Once Per Combat - this table is destroyed/reset on enter/exit combat
	}

	self:InitOptionsGUIState() -- sets up event lists and current selections
end


function SpeakinSpell:DeleteSpell(key)
	-- delete the current selection from the SavedData and RuntimeData tables
	SpeakinSpellSavedData.EventTable[key] = nil
	
	if not self.RuntimeData then
		-- we're still initializing, so the rest of this stuff doesn't matter
		return
	end
	
	-- remove runtime data for messaging cooldowns and limits
	self.RuntimeData.AnnouncementHistory[key] = nil
	
	-- update GUI controls
	self.OptionsGUI.args.CurrentMessagesGUI.args.EventSelect.values[key] = nil
	
	-- update selected item
	self.OptionsGUIStates.MessageSettings.SelectedEventKey = nil
	
	-- restore the spell to the list of new detected
	-- so the user doesn't have to retrigger it
	-- self:RecordNewSpell(EventTableEntry.DetectedEvent)
	-- NO! could be outdated and bugged event
	-- just make the user recreate it	
	
	self:CurrentMessagesGUI_ValidateSelectedEvent() -- make sure we still have a valid spell selection
end



function SpeakinSpell:IsEventTypeRankable(checktype)
	local RankableEventTypes = {
		"ALL",
		"UNIT_SPELLCAST_SENT",
		"UNIT_SPELLCAST_CHANNEL_START",
		"UNIT_SPELLCAST_CHANNEL_STOP",
		"UNIT_SPELLCAST_FAILED",
		"UNIT_SPELLCAST_INTERRUPTED",
		"UNIT_SPELLCAST_STOP",
		"UNIT_SPELLCAST_SUCCEEDED",
	}
	for _,type in pairs(RankableEventTypes) do
		if checktype == type then
			return true --this event type is rankable
		end
	end
	
	return false --this event type is not rankable
end


function SpeakinSpell:IsSelectedEventTypeFilterRankable()
	return self:IsEventTypeRankable( self.RuntimeData.SelectedEventTypeFilter )
end


function SpeakinSpell:SetFilterShowAllRanks(value)
	if SpeakinSpellSavedData.ShowAllRanks == value then
		--value is not changing
		return
	end
	
	SpeakinSpellSavedData.ShowAllRanks = value
	
	if self:IsSelectedEventTypeFilterRankable() then
		--self.OptionsGUIStates.MessageSettings.FilterChanged = true --the message settings always show all ranks
		self.OptionsGUIStates.CreateNew.FilterChanged = true
	end
end


function SpeakinSpell:SetFilter(type,text)
	local rebuild = false
	if type and self.RuntimeData.SelectedEventTypeFilter ~= type then
		self.RuntimeData.SelectedEventTypeFilter = type
		rebuild = true
	end
	
	if text and self.RuntimeData.SelectedEventTextFilter ~= text then
		self.RuntimeData.SelectedEventTextFilter = text
		rebuild = true
	end
	
	if self.OptionsGUIStates and rebuild then
		self.OptionsGUIStates.MessageSettings.FilterChanged = true
		self.OptionsGUIStates.CreateNew.FilterChanged = true
	end
end

	
function SpeakinSpell:MatchesFilter(DetectedEvent, ForceAllRanks)
	--abbreviations
	local type = SpeakinSpell.RuntimeData.SelectedEventTypeFilter
	local text = SpeakinSpell.RuntimeData.SelectedEventTextFilter
	
	-- check type filter
	if type and not (type == "ALL" or type == DetectedEvent.type) then
		return false
	end
	
	-- check text filter - use lowercase for case-insensitive comparison
	if text and not (text == "") then
		local ndxMatch = string.find(string.lower(DetectedEvent.name), string.lower(text))
		if ndxMatch and ndxMatch >= 0 then
			return true
		else
			return false
		end
	end
	
	-- check rank filter
	if self:IsSelectedEventTypeFilterRankable() then
		local HideSpecialRanks = not (ForceAllRanks or SpeakinSpellSavedData.ShowAllRanks)
		if HideSpecialRanks then
			if DetectedEvent.rank and (DetectedEvent.rank ~= "") then
				-- a rank is specified
				return false
			end
		end
	end
	
	-- no filters are defined
	return true
end


-- RecordNewEvent creates and maintains a list of new spells that don't have messages associated with them
-- this list is used to present the user with a list of options when she wants to start announcing a new spell
function SpeakinSpell:RecordNewEvent(DetectedEvent)
	local funcname = "RecordNewEvent"
	
	-- if we already recorded this spell, don't add a duplicate, unless this is a different rank
	-- NOTE: RecordNewEvent is called for both ranked, and rankless versions of the event
	if not SpeakinSpellSavedData.NewEventsDetected[DetectedEvent.key] then
		self:DebugMsg(funcname, DetectedEvent.SpellLink)
		SpeakinSpellSavedData.NewEventsDetected[DetectedEvent.key] = DetectedEvent
		-- update the options GUI if it exists
		if self.OptionsGUI then
			if self:MatchesFilter(DetectedEvent,false) then
				-- add this to the list
				self.OptionsGUI.args.CreateNew.args.NewSpellNameSelect.values[DetectedEvent.key] = DetectedEvent.DisplayName
				-- re sort the list
				table.sort(self.OptionsGUI.args.CreateNew.args.NewSpellNameSelect.values)
				-- make this new item the selection
				self.OptionsGUIStates.CreateNew.SelectedEventKey = DetectedEvent.key
			end
		end
	end
end


function SpeakinSpell:FormatDisplayName(de)
	-- create a display name for this event
	if de.rank ~= "" then
		return L.EVENTTYPES.IN_SPELL_LIST[de.type]..de.name.." ("..de.rank..")"
	else
		if self:IsEventTypeRankable( de.type ) then
			return L.EVENTTYPES.IN_SPELL_LIST[de.type]..de.name.." "..L["(Any Rank)"]
		else --other supported events (buffs, debuffs, macros, and system events) donever 't have ranks
			return L.EVENTTYPES.IN_SPELL_LIST[de.type]..de.name
		end
	end
end


function SpeakinSpell:SafeGetSpellLink(de)
	local SpellLink,_ = GetSpellLink(de.name, de.rank)
	if SpellLink then
		return SpellLink
	else
		return "["..de.name.."]"
	end
end



-- returns rankedkey, ranklesskey
-- if not a rankable event, then ranklesskey will be nil
-- TODOFUTURE: this could return an ordered list of potential keys?
-- maybe for linking UNIT_SPELLCAST_SENT to the matching SPELL_AURA_GAIN 
-- and/or for making speech sets that apply to many spells with the same speech settings
function SpeakinSpell:MakeEventTableKeys( de )
	-- create a unique key for this event
	local rankedkey		= string.gsub( string.upper(de.type..de.name..de.rank), " ", "_" ) -- ACTIVE 3.1.2.07 format
	local ranklesskey	= string.gsub( string.upper(de.type..de.name),			" ", "_" ) -- ACTIVE 3.1.2.07 format
	
	if self:IsEventTypeRankable( de.type ) and (rankedkey ~= ranklesskey) then
		return rankedkey, ranklesskey
	else
		return rankedkey, nil
	end
end


function SpeakinSpell:CreateDetectedEvent( DetectedEventStub )
	local funcname = "CreateDetectedEvent"
	
	local de = self:CopyTable(DetectedEventStub)
	
	if not de.rank then
		de.rank = ""
	end
	
	if not de.caster then
		de.caster = UnitName("player")
	end
	
	-- create a spell link, if possible
	-- the stub will set SpellLink = true if the event is a spell
	-- else will we make a fake one
	de.SpellLink = self:SafeGetSpellLink(de)
	
	-- validate the target - get a backup target from selection, focus, or assumed self-cast
	if not de.target or de.target == "" or de.target == "Unknown" then
		de.target = self:GetDefaultTarget(true)
	end
	
	-- fetch additional target information
	de.targetclass = UnitClass(de.target)
	de.targetrace = UnitRace(de.target)

	-- create a display name for this event
	de.DisplayName = self:FormatDisplayName(de)
	
	return de
end


-------------------------------------------------------------------------------
-- CHECK OPTIONS IF SPEAKING IS ALLOWED
-------------------------------------------------------------------------------


function SpeakinSpell:SaveAnnouncementHistory( de, msg )
	-- record info about the last time we announced this event (which should be right now)
	self.RuntimeData.AnnouncementHistory[de.key] = {
		LastMessage = msg,
		LastTime	= GetTime(),
		LastTarget	= de.target,
	}
	self.RuntimeData.AnnouncedThisCombat[de.key] = true
end


function SpeakinSpell:CheckSpellFrequency( de )
	-- check the frequency setting against a random number to determine if we should speak or not
	local Chance = math.random(1,100)
	local Frequency = de.EventTableEntry.Frequency*100

	if Chance > Frequency then
		self:DebugMsg( "CheckSpellFrequency", string.format("random chance failed for '%s' rolled %d > %d", de.SpellLink, Chance, Frequency) )
		return false
	end
	
	return true
end


function SpeakinSpell:CheckSpellCooldown( de )
	local funcname = "CheckSpellCooldown"
	
	-- check for a cooldown limit on announcing this spell

	local History = self.RuntimeData.AnnouncementHistory[de.key]
	if not History then
		-- we have not yet announced this spell, so it can't be on cooldown
		return true
	end

	-- calculate elapsed time since last announcement
	-- NOTE: History.LastTime is guaranteed to be a valid value, or we would have aborted above
	local elapsed = GetTime() - History.LastTime

	-- Now finally actually check if the cooldown is in effect
	-- NOTE: "no cooldown" is set by Cooldown=0, so elapsed < 0 will not occur
	if elapsed < de.EventTableEntry.Cooldown then
		-- the cooldown is in effect, so be silent
		self:DebugMsg( funcname, string.format("messaging on cooldown: %s: elapsed:%d, CD:%d", de.SpellLink, elapsed, de.EventTableEntry.Cooldown) )
		return false
	end
	
	return true
end


function SpeakinSpell:CheckOncePerTarget( de )
	local funcname = "CheckOncePerTarget"

	if not de.EventTableEntry.OncePerTarget then
		-- "limit once per target" is disabled, so we can always speak
		return true
	end

	local History = self.RuntimeData.AnnouncementHistory[de.key]
	if not History then
		-- we haven't announced this event yet at all, so we can't have announced it for this target
		return true
	end
	
	-- check the last target that was used
	if de.target == History.LastTarget then
		self:DebugMsg( funcname, "Limit once per target is in effect for target="..tostring(de.target) )
		return false
	end
	
	-- we have a new target this time, so we can speak
	return true
end


function SpeakinSpell:CheckOncePerCombat( de )
	local funcname = "CheckOncePerCombat"

	if not de.EventTableEntry.OncePerCombat then
		-- "limit once per combat" is disabled, so we can always speak
		return true
	end
	
	local History = self.RuntimeData.AnnouncementHistory[de.key]
	if not History then
		-- we haven't announced this event yet at all, so we can't have announced it for this target
		return true
	end

--	if not self.RuntimeData.InCombat then
--		-- we're not in combat, so this option should not apply
--		self:DebugMsg( funcname, "Limit once per combat no has no effect because you are not in combat" )
--		return true
--	end
	
	if self.RuntimeData.AnnouncedThisCombat[de.key] then
		self:DebugMsg( funcname, "Limit once per combat is in effect for this event key="..tostring(de.key) )
		return false
	end
	
	-- it's a new combat, so go ahead and announce
	return true
end


function SpeakinSpell:ResetOncePerCombatFlags()
	-- forget all the once-per-combat memory
	-- NOTE: this is easier if we track it as a separate table
	--		rather than add data to self.RuntimeData.AnnouncementHistory
	self.RuntimeData.AnnouncedThisCombat = {}
	
	-- NOTE: don't reset last target info when changing combat sessions
	-- because we want "limit once per target" to apply to targets to the same name in multiple pulls
end


function SpeakinSpell:AllowSpeakForSpell( de )
	local funcname = "AllowSpeakForSpell"
	
	if de.EventTableEntry.DisableAnnouncements then
		self:DebugMsg(funcname,"OnEventDetected - messages for " .. de.SpellLink .. " are disabled")
		return false
	end
	
	if not SpeakinSpellSavedData.EnableAllMessages then
		self:DebugMsg(funcname,"OnEventDetected - all SpeakinSpell messages are disabled")
		return false
	end
	
	if not self:CheckSpellFrequency( de ) then
		return false
	end
	
	if not self:CheckSpellCooldown( de ) then
		return false
	end
	
	if not self:CheckOncePerTarget( de ) then
		return false
	end
	
	if not self:CheckOncePerCombat( de ) then
		return false
	end
	
	-- no limits or restrictions apply, so go ahead and speak
	return true
end

